# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import config
import log
import window
import events, hub
import statusbar
import encoding

from helper import formattime

class playerwin(window.window):
    def __init__(self, screen, layout, channel, playerid):
        self.song = None
        self.time = 0
        self.paused = 0
        self.stopped = 1
        self.playerid = playerid
        self.keybindings = config.keybindings.general
        self.songformat = config.playerwindow.songformat
        self.playerinfofile = config.general.playerinfofile
        self.songchangecommand = config.general.songchangecommand.strip()
        if self.songchangecommand:
            self.songchangecommand = "( %s )&" % self.songchangecommand
        h, w, y, x, border = layout
        window.window.__init__(self, screen, h, w, y, x,
                               config.colors.playerwindow,
                               _("Playback Info"), border)
	try:
            self.playerinfofd = open(self.playerinfofile, "w")
        except IOError, e:
            log.error(_("error '%s' occured during write to playerinfofile") % e)
            self.playerinfofd = None
        # we don't want to have the focus
        self.bottom()

        channel.subscribe(events.playbackinfochanged, self.playbackinfochanged)
        channel.subscribe(events.keypressed, self.keypressed)
        self.update()

    def resize(self, layout):
        h, w, y, x, self.border = layout
        window.window.resize(self, h, w, y, x)

    def updatestatusbar(self):
        if self.song and not self.paused and self.keybindings["playerpause"]:
            sbar = statusbar.generatedescription("general", "playerpause")
        else:
            sbar = statusbar.generatedescription("general", "playerstart")
        hub.notify(events.statusbar_update(1, sbar))
        
    def update(self):
        window.window.update(self)
        self.updatestatusbar()
        self.addstr(1, self.ix, " "*self.iw)
        if self.song:
            self.move(1, self.ix)
            s1 = _("Time:")
            s2 = " %s/%s " % (formattime(self.time), formattime(self.song.length))
            self.addstr(s1, self.colors.description)
            self.addstr(s2, self.colors.content)

            if not self.paused:
                barlen = self.iw-len(s1)-len(s2)
                try:
                    percentplayed = int(barlen*self.time/self.song.length)
                except ZeroDivisionError:
                    percentplayed = 0
                self.addstr("#"*(percentplayed), self.colors.progressbarhigh)
                self.addstr("#"*(barlen-percentplayed), self.colors.progressbar)
            else:
                self.addstr(_("paused"), self.colors.description)

    # event handler

    def playbackinfochanged(self, event):
        if event.playbackinfo.playerid == self.playerid:
            if self.song != event.playbackinfo.song and event.playbackinfo.song and self.songchangecommand:
                os.system(event.playbackinfo.song.format(self.songchangecommand, safe=True))
            self.song = event.playbackinfo.song
            self.paused = event.playbackinfo.ispaused()
            self.stopped = event.playbackinfo.isstopped()
            if self.song:
                self.settitle(u"%s%s" % (event.playbackinfo.iscrossfading() and "-> " or "", self.song.format(self.songformat)))
            else:
                self.settitle(_("Playback Info"))
            self.time = event.playbackinfo.time
            self.update()

            # update player info file, if configured
            if self.playerinfofd:
                try:
                    self.playerinfofd.seek(0)
                    if self.song:
                        info = "%s - %s (%s/%s)\n"  % ( self.song.artist,
                                                        self.song.title,
                                                        formattime(self.time),
                                                        formattime(self.song.length))
                    else:
                        info = _("Not playing") + "\n"
                    info = encoding.encode(info)
                    self.playerinfofd.write(info)
                    self.playerinfofd.truncate(len(info))
                except IOError, e:
                    log.error(_("error '%s' occured during write to playerinfofile") % e)
                    self.playerinfofd = None

    def keypressed(self, event):
        key = event.key
        if key in self.keybindings["playerstart"] and self.paused:
            hub.notify(events.playerstart(self.playerid))
        elif key in self.keybindings["playerpause"] and not self.paused and not self.stopped:
            hub.notify(events.playerpause(self.playerid))
        elif key in self.keybindings["playerstart"]:
            hub.notify(events.playerstart(self.playerid))
        elif key in self.keybindings["playernextsong"]:
            hub.notify(events.playernext(self.playerid))
        elif key in self.keybindings["playerprevioussong"]:
            hub.notify(events.playerprevious(self.playerid))
        elif key in self.keybindings["playerrewind"]:
            hub.notify(events.playerseekrelative(self.playerid, -2))
        elif key in self.keybindings["playerforward"]:
            hub.notify(events.playerseekrelative(self.playerid, 2))
        elif key in self.keybindings["playerstop"]:
            hub.notify(events.playerstop(self.playerid))
        elif key in self.keybindings["playerplayfaster"]:
            hub.notify(events.playerplayfaster(self.playerid))
        elif key in self.keybindings["playerplayslower"]:
            hub.notify(events.playerplayslower(self.playerid))
        elif key in self.keybindings["playerspeedreset"]:
            hub.notify(events.playerspeedreset(self.playerid))
        elif key in self.keybindings["playerratecurrentsong1"]:
            if self.song:
                self.song.rate(1)
        elif key in self.keybindings["playerratecurrentsong2"]:
            if self.song:
                self.song.rate(2)
        elif key in self.keybindings["playerratecurrentsong3"]:
            if self.song:
                self.song.rate(3)
        elif key in self.keybindings["playerratecurrentsong4"]:
            if self.song:
                self.song.rate(4)
        elif key in self.keybindings["playerratecurrentsong5"]:
            if self.song:
                self.song.rate(5)
        else:
            return
        raise hub.TerminateEventProcessing




# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import curses
import config
import hub, events
import help
import window
import events
import encoding

# indicator which can be used to signalise that the statusbar
# has to be terminated here
terminate = [("TERMINATE", 0)]

def generatedescription(section, name):
    primarykey = config.keybindings[section][name][0]
    keyname = help.getkeyname(primarykey)

    # save some space
    if keyname.startswith("<") and keyname.endswith(">"):
        keyname = keyname[1:-1]

    return [(keyname, config.colors.statusbar.key),
            (":"+help.descriptions[section][name][0],
             config.colors.statusbar.description)]


class statusbar(window.window):
    def __init__(self, screen, line, width, channel):
        self.channel = channel
        # self.content[0]: local info
        # self.content[1]: player info
        # self.content[2]: global info
        self.content = [[], [], []]

        # message overriding contents
        self.message = None
        # for identification purposes, we only generate this once
        self.removemessageevent = events.statusbar_showmessage(None)
        window.window.__init__(self, screen,
                               1, width, line, 0,
                               config.colors.statusbar,
                               None)
        self.channel.subscribe(events.statusbar_update, self.statusbar_update)
        self.channel.subscribe(events.statusbar_showmessage, self.statusbar_showmessage)

        # hack to export some properties of the statusbar singleton into
        # the module namespace
        global separator
        if width<=80:
            separator = [(" ", self.colors.background)]
        else:
            separator = [("  ", self.colors.background)]


    def resize(self, line, width):
        window.window.resize(self, 1, width, line, 0)

        global separator
        if width<=80:
            separator = [(" ", self.colors.background)]
        else:
            separator = [("  ", self.colors.background)]

    # event handler

    def statusbar_update(self, event):
        self.content[event.pos] = event.content
        self.update()

    def statusbar_showmessage(self, event):
        self.message = event.message
        if self.message:
            # make message disappear after a certain time
            hub.notify(events.sendeventin(self.removemessageevent, 2, replace=1))
        self.update()

        # we want to get a message out immediately, so we force its output
        curses.panel.update_panels()
        curses.doupdate()

    # update method

    def update(self):
        window.window.update(self)
        self.move(0,0)
        self.clrtoeol()

        if self.message:
            self.addstr(encoding.encode(self.message), config.colors.statusbar.key)
        else:
            for element in self.content[0]+separator+self.content[1]+separator+self.content[2]:
                if element==terminate[0]:
                    break
                self.addstr(encoding.encode(element[0]), element[1])



# -*- coding: ISO-8859-1 -*-

import events
import plugin
import config
import log
import copy

# We define a config section, which will be called [plugin.minimal]
# In the plugin, we will be able to access the values as member variable
# self.config

class config(config.configsection):
    message = config.configstring("new song: ")

class plugin(plugin.plugin):

    """ a simple plugin that logs when a new song is played on the main player

    A configuration option message allows the user to specify the notification string.
    """

    def start(self):
        # its good practice to notify the user of the started plugin
        log.info("started minimal plugin")
        self.playbackinfo = None
        self.channel.subscribe(events.playbackinfochanged, self.playbackinfochanged)

    # event handler

    def playbackinfochanged(self, event):
        if event.playbackinfo is None:
            return
        if self.playbackinfo is None or self.playbackinfo.song != event.playbackinfo.song:
            log.info("%s%s" % (self.config.message, event.playbackinfo.song))
            self.playbackinfo = copy.copy(event.playbackinfo)




# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004, 2005, 2007 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

# uncomment next line for pychecker
# import gettext; gettext.install("PyTone", "")

import curses, curses.panel, fcntl, os, struct, termios
import events, hub
import filelistwin
import playlistwin
import playerwin
import iteminfowin
import lyricswin
import helpwin
import inputwin
import mixerwin
import log
import logwin
import statswin
import statusbar
import config


class mainscreen:
    def __init__(self, screen, songdbids, playerids, plugins):
        self.screen = screen
        self.layout = config.general.layout
        self.h, self.w = self.getmaxyx()
        log.debug("h=%d, w=%d" % (self.h, self.w))
        self.channel = hub.newchannel()
        self.keybindings = config.keybindings.general
        self.done = False

        self.statusbar = statusbar.statusbar(screen, self.h-1, self.w, self.channel)

        # first we setup the input window in order to have it as first window
        # in the keypressed events lists.
        if config.inputwindow.type=="popup":
            self.inputwin = inputwin.popupinputwin(screen, self.h, self.w, self.channel)
        else:
            self.inputwin = inputwin.statusbarinputwin(screen, self.h, self.w, self.channel)

        # setup the four main windows
        windowslayout = self.calclayout()
        self.playerwin = playerwin.playerwin(screen, windowslayout["playerwin"], self.channel, playerids[0])
        self.iteminfowin = iteminfowin.iteminfowin(screen, windowslayout["iteminfowin"], self.channel, playerids, playerids[1])
        self.filelistwin = filelistwin.filelistwin(screen, windowslayout["filelistwin"], self.channel, songdbids)
        self.playlistwin = playlistwin.playlistwin(screen, windowslayout["playlistwin"], self.channel, "main")
        self.connectborders()

        # setup additional windows which appear on demand
        self.helpwin = helpwin.helpwin(screen, self.h, self.w, self.channel)

        self.logwin = logwin.logwin(screen, self.h, self.w, self.channel)
        self.statswin = statswin.statswin(screen, self.h, self.w, self.channel, len(songdbids))

        self.iteminfowinlong = iteminfowin.iteminfowinlong(screen, self.h, self.w, self.channel)
        self.lyricswin = lyricswin.lyricswin(screen, self.h, self.w, self.channel)

        self.mixerwin = None
        if config.mixer.device:
            try:
                if config.mixerwindow.type=="popup":
                    self.mixerwin = mixerwin.popupmixerwin(screen, self.h, self.w, self.channel)
                else:
                    self.mixerwin = mixerwin.statusbarmixerwin(screen, self.h, self.w, self.channel)

            except IOError, e:
                log.warning('error "%s" during mixer init - disabling mixer' % e)
        else:
            # disbable keybindings to obtain correct help window contents
            del config.keybindings.general.volumeup
            del config.keybindings.general.volumedown

        # now we start the plugins
        for pluginmodule, pluginconfig in plugins:
            plugin_class = pluginmodule.plugin
            if plugin_class:
                plugin = plugin_class(self.channel, pluginconfig, self)
                plugin.start()

        self.channel.subscribe(events.keypressed, self.keypressed)
        self.channel.subscribe(events.activateplaylist, self.activateplaylist)
        self.channel.subscribe(events.activatefilelist, self.activatefilelist)
        self.channel.subscribe(events.quit, self.quit)

        hub.notify(events.activatefilelist())

    def run(self):
        """ main loop of control thread """
        skipcount = 0
        while not self.done:
            try:
                key = self.screen.getch()

                if key==27:
                    # handle escape sequence (e.g. alt+key)
                    key = self.screen.getch()+1024

                if key==curses.KEY_MOUSE:
                    mouse = curses.getmouse()
                    x, y = mouse[1:3]
                    state = mouse[4]
                    hub.notify(events.mouseevent(y, x, state))
                elif key==curses.KEY_RESIZE:
                    self.resizeterminal()
                elif key in self.keybindings["exit"]:
                    curses.halfdelay(5)
                    key = self.screen.getch()
                    curses.halfdelay(1)
                    if key in self.keybindings["exit"]:
                        break
                elif key!=-1:
                    hub.notify(events.keypressed(key), 50)

                self.channel.process()

                # update screen
                if key == -1 or skipcount >= config.general.throttleoutput:
                    curses.panel.update_panels()
                    curses.doupdate()
                    skipcount = 0
                else:
                    skipcount += 1

            except KeyboardInterrupt:
                pass

    def quit(self, event):
        """ cleanup """
        self.done = True

    def getmaxyx(self):
        # taken from http://dag.wieers.com/home-made/dstat/
        try:
            h, w = int(os.environ["LINES"]), int(os.environ["COLUMNS"])
        except KeyError:
            try:
                h, w = curses.tigetnum('lines'), curses.tigetnum('cols')
            except:
                try:
                    s = struct.pack('HHHH', 0, 0, 0, 0)
                    x = fcntl.ioctl(sys.stdout.fileno(), termios.TIOCGWINSZ, s)
                    h, w = struct.unpack('HHHH', x)[:2]
                except:
                    h, w = 25, 80

        # take into account minimal height and width according to layout
        if self.layout == "onecolumn":
            minh = 27
            minw = 25
        else:
            minh = 17
            minw = 65
        return max(h, minh), max(w, minw)

    def calclayout(self):
        """ calculate layout of four main windows for given height h and width w of mainscreen and
        layout type layout"""
        result = {}
        if self.layout == "twocolumn":
            leftpanelw = int(self.w/2.2)
            rightpanelw = self.w-leftpanelw
            rightpanelx = leftpanelw

            if config.playerwindow.border == config.BORDER_COMPACT:
                playerwinb = config.BORDER_LEFT | config.BORDER_TOP | config.BORDER_RIGHT
            elif config.playerwindow.border == config.BORDER_ULTRACOMPACT:
                playerwinb = config.BORDER_LEFT | config.BORDER_TOP
            else:
                playerwinb = config.playerwindow.border
            if playerwinb & config.BORDER_BOTTOM:
                playerwinh = 3
            else:
                playerwinh = 2
            result["playerwin"] = playerwinh, rightpanelw, 0, rightpanelx, playerwinb

            if config.iteminfowindow.border == config.BORDER_COMPACT:
                iteminfowinb = config.BORDER_LEFT | config.BORDER_TOP | config.BORDER_RIGHT
            elif config.iteminfowindow.border == config.BORDER_ULTRACOMPACT:
                iteminfowinb = config.BORDER_LEFT | config.BORDER_TOP
            else:
                iteminfowinb = config.iteminfowindow.border
            if iteminfowinb & config.BORDER_BOTTOM:
                iteminfowinh = 6
            else:
                iteminfowinh = 5
            result["iteminfowin"] = iteminfowinh, rightpanelw, playerwinh, rightpanelx, iteminfowinb

            if config.filelistwindow.border == config.BORDER_COMPACT:
                filelistwinb = config.BORDER_LEFT | config.BORDER_TOP | config.BORDER_BOTTOM
            elif config.filelistwindow.border == config.BORDER_ULTRACOMPACT:
                filelistwinb = config.BORDER_TOP
            else:
                filelistwinb = config.filelistwindow.border
            result["filelistwin"] = self.h-1, leftpanelw, 0, 0, filelistwinb
            
            if config.playlistwindow.border == config.BORDER_COMPACT:
                playlistwinb = config.BORDER_LEFT | config.BORDER_TOP | config.BORDER_RIGHT | config.BORDER_BOTTOM
            elif config.playlistwindow.border == config.BORDER_ULTRACOMPACT:
                playlistwinb = config.BORDER_LEFT | config.BORDER_TOP
            else:
                playlistwinb = config.playlistwindow.border
            result["playlistwin"] = (self.h-iteminfowinh-playerwinh-1, rightpanelw, iteminfowinh+playerwinh, rightpanelx,
                                     playlistwinb)

        else:
            # onecolumn layout
            if config.playerwindow.border == config.BORDER_COMPACT:
                playerwinb = config.BORDER_LEFT | config.BORDER_TOP | config.BORDER_RIGHT
            elif config.playerwindow.border == config.BORDER_ULTRACOMPACT:
                playerwinb = config.BORDER_TOP
            else:
                playerwinb = config.playerwindow.border
            if playerwinb & config.BORDER_BOTTOM:
                playerwinh = 3
            else:
                playerwinh = 2
            result["playerwin"] = playerwinh, self.w, 0, 0, playerwinb

            if config.iteminfowindow.border == config.BORDER_COMPACT:
                iteminfowinb = config.BORDER_LEFT | config.BORDER_TOP | config.BORDER_RIGHT
            elif config.iteminfowindow.border == config.BORDER_ULTRACOMPACT:
                iteminfowinb = config.BORDER_TOP
            else:
                iteminfowinb = config.iteminfowindow.border
            if iteminfowinb & config.BORDER_BOTTOM:
                iteminfowinh = 6
            else:
                iteminfowinh = 5
            result["iteminfowin"] = iteminfowinh, self.w, playerwinh, 0, iteminfowinb

            if config.filelistwindow.border == config.BORDER_COMPACT:
                filelistwinb = config.BORDER_LEFT | config.BORDER_TOP | config.BORDER_RIGHT
            elif config.filelistwindow.border == config.BORDER_ULTRACOMPACT:
                filelistwinb = config.BORDER_TOP
            else:
                filelistwinb = config.filelistwindow.border
            filelistwinh = int(self.h/2)
            result["filelistwin"] = filelistwinh, self.w, playerwinh+iteminfowinh, 0, filelistwinb

            if config.playlistwindow.border == config.BORDER_COMPACT:
                playlistwinb = config.BORDER_LEFT | config.BORDER_TOP | config.BORDER_RIGHT | config.BORDER_BOTTOM
            elif config.playlistwindow.border == config.BORDER_ULTRACOMPACT:
                playlistwinb = config.BORDER_TOP
            else:
                playlistwinb = config.playlistwindow.border
            result["playlistwin"] = (self.h-iteminfowinh-playerwinh-filelistwinh-1, self.w, playerwinh+iteminfowinh+filelistwinh, 0,
                                     playlistwinb)

        return result

    def connectborders(self):
        borderends = []
        mainwindows = [self.filelistwin, self.playerwin, self.iteminfowin, self.playlistwin]
        for win in mainwindows:
            borderends.extend(win.getborderends())

        for win in mainwindows:
            win.connectborderends(borderends)
            win.update()

    def refresh(self):
        # refresh screen (i don't know any better way...)
        self.screen.clear()
        self.filelistwin.update()
        self.playlistwin.update()
        self.playerwin.update()
        self.iteminfowin.update()
        self.statusbar.update()

    def resizeterminal(self):
        self.h, self.w = self.getmaxyx()

        windows = self.calclayout()
        self.filelistwin.resize(windows["filelistwin"])
        self.playerwin.resize(windows["playerwin"])
        self.iteminfowin.resize(windows["iteminfowin"])
        self.playlistwin.resize(windows["playlistwin"])
            
        self.statusbar.resize(self.h-1, self.w)
        self.helpwin.resize(self.h, self.w)
        self.logwin.resize(self.h, self.w)
        self.statswin.resize(self.h, self.w)
        self.inputwin.resize(self.h, self.w)

        if self.mixerwin:
            self.mixerwin.resize(self.h, self.w)

        self.connectborders()

        self.filelistwin.update()
        self.playlistwin.update()
        self.playerwin.update()
        self.iteminfowin.update()
        self.statusbar.update()

    # event handler

    def keypressed(self, event):
        key = event.key
        if key in self.keybindings["refresh"]:
            self.refresh()
        elif key in self.keybindings["playlistdeleteplayedsongs"]:
            hub.notify(events.playlistdeleteplayedsongs())
        elif key in self.keybindings["playlistclear"]:
            hub.notify(events.playlistclear())
            hub.notify(events.activatefilelist())
        elif key in self.keybindings["playlistreplay"]:
            hub.notify(events.playlistreplay())
        elif key in self.keybindings["playlistsave"]:
            hub.notify(events.playlistsave())
        elif key in self.keybindings["playlisttoggleautoplaymode"]:
            hub.notify(events.playlisttoggleautoplaymode())
        elif key in self.keybindings["showhelp"]:
            if self.filelistwin.hasfocus():
                context = "filelistwindow"
            elif self.playlistwin.hasfocus():
                context = "playlistwindow"
            else:
                context = None
            self.helpwin.showhelp(context)
        elif key in self.keybindings["showlog"]:
            self.logwin.show()
        elif key in self.keybindings["showstats"]:
            self.statswin.show()
        elif key in self.keybindings["showiteminfolong"]:
            self.iteminfowinlong.show()
        elif key in self.keybindings["showlyrics"]:
            self.lyricswin.show()
        elif key in self.keybindings["togglelayout"]:
            self.layout = self.layout == "onecolumn" and "twocolumn" or "onecolumn"
            self.resizeterminal()
        else:
            log.debug("unknown key: %d" % key)

    def activateplaylist(self, event):
        self.playlistwin.top()

    def activatefilelist(self, event):
        self.filelistwin.top()



##############################################################################
#
# Copyright (c) 2004 TINY SPRL. (http://tiny.be) All Rights Reserved.
#                    Fabien Pinckaers <fp@tiny.Be>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
import os.path
import urllib
import md5
import time
import csv

class BadAuthError(Exception):
    pass

class NotConnectedError(Exception):
    pass

class SubmissionError(Exception):
    pass

class Scrobbler(object):

    client = 'tst'
    version = '0.1'
    url = 'http://post.audioscrobbler.com/'
    backlogfile = os.path.expanduser('~/.audioscrobbler')

    def __init__(self, user, password):
        self.user = user
        self.password = password
        self.connected = False
        self.lastconnected = 0

    def dohandshake(self):
        timestamp = str(int(time.time()))
        pswd = md5.new(md5.new(self.password).hexdigest() +
                timestamp).hexdigest()
        rs = urllib.urlencode(
                { 'hs' : 'true',
                  'p' : '1.1',
                  'c' : self.client,
                  'v' : self.version,
                  'u' : self.user,
                  't' : int(time.time()),
                  'a' : pswd })
        url = self.url + "?" + rs
        result = urllib.urlopen(url).readlines()
        if result[0].startswith('UPTODATE'):
            return True, False, result[1:]
        elif result[0].startswith('UPDATE'):
            return True, True, result[1:]
        elif result[0].startswith('FAILED'):
            return False, True, result
        elif result[0].startswith('BADUSER'):
            return False, False, result

    def handshake(self):
        connected = False
        nbtries = 0
        while not connected:
            connected, newclient, results = self.dohandshake()
            if connected:
                self.connected = True
                self.newclient = newclient
                self.md5 = results[0][:-1]
                self.submitURL = results[1][:-1]
                self.lastconnected = time.time()
            else:
                if not newclient:
                    connected = True
                    raise BadAuthError
                time.sleep(2**nbtries * 60)
                nbtries += 1
                if nbtries > 4:
                    connected = True

    def submit(self, song):
        self.sendInfo(song.artist, song.title, song.album, song.length,
                             song.date_lastplayed)

    def sendInfo(self, artist, title, album, length, debut):
        if self.lastconnected - time.time() > 300:
            self.connected = False
            self.newclient = False
            self.md5 = ''
            self.submitURL = ''
            self.handshake()

        if not self.connected:
            raise NotConnectedError

        pswd = md5.new(md5.new(self.password).hexdigest() +
                self.md5).hexdigest()
        infodict = { 'u' : self.user, 's' : pswd, 'a[0]' : artist,
            't[0]' : title, 'b[0]' : album, 'm[0]' : '', 'l[0]' : length,
            'i[0]' : time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(debut))}
        rs = urllib.urlencode(infodict)

        try:
            result = urllib.urlopen(self.submitURL, rs).readlines()
        except IOError:
            self.addBacklog(infodict)
            return False

        if result[0].startswith('OK'):
            return True
        elif result[0].startswith('FAILED'):
            raise SubmissionError
        elif result[0].startswith('BADAUTH'):
            raise BadAuthError

    def addBacklog(self, info):
        fd = file(self.backlogfile, 'wa')
        writer = csv.writer(fd)
        writer.writerow([info["a[0]"], info["t[0]"], info["b[0]"], info["l[0]"],
                info["i[0]"]])
        fd.close()

    def getBacklog(self):
        try:
            reader = csv.reader(file(self.backlogfile, 'r'))
        except IOError:
            return
        backlogs = list(reader)
        file(self.backlogfile, 'w').close()
        for entry in backlogs:
            self.sendInfo(*entry)
            time.sleep(1)

def main():
    class A:
        pass
    a = A()
    a.artist = raw_input('Artiste: ')
    a.title = raw_input('Titre: ')
    a.album = raw_input('Album: ')
    a.length = raw_input('Duree: ')
    a.lastplayed = [ int(time.time() - 250) ]

    scrobbler = Scrobbler('user', '*****')
    scrobbler.handshake()
    scrobbler.submit(a)

if __name__ == '__main__':
    main()



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import config
import curses
import string
import events, hub
import statusbar
import window

# string.printable is not updated when locale is changed (this is a known bug, which
# however is not planed to be fixed), so we just do this by ourselves

printable = string.digits + string.letters + string.punctuation + string.whitespace

class inputwin(window.window):

    """ generic input window """

    def __init__(self, screen, maxh, maxw, channel):
        self.channel = channel
        self.keybindings = config.keybindings.general
        self.inputstring = ""
        self.inputprompt = ""

        self.hide()

        self.channel.subscribe(events.keypressed, self.keypressed)
        self.channel.subscribe(events.mouseevent, self.mouseevent)
        self.channel.subscribe(events.requestinput, self.requestinput)

    # we also need a blinking cursor, whenever we have the focus

    def hide(self):
        try:
            curses.curs_set(0)
        except:
            pass
        window.window.hide(self)

    def top(self):
        try:
            curses.curs_set(1)
        except:
            pass
        window.window.top(self)
        
    # event handler

    def keypressed(self, event):
        if self.hasfocus():
            key = event.key
            if 32 <= key <= 255 and chr(key) in printable:
                if len(self.inputstring)+len(self.inputprompt)<self.iw-1:
                    self.inputstring += chr(key)
                else:
                    self.inputstring = self.inputstring[:-1]+chr(key)
            elif key == ord("\n"):
                self.hide()
            elif key == 1023:
                self.hide()
            elif key == curses.KEY_BACKSPACE:
                self.inputstring = self.inputstring[:-1]
                
            self.update()
            self.inputhandler(self.inputstring, key)
            raise hub.TerminateEventProcessing

    def mouseevent(self, event):
        if self.hasfocus():
            self.hide()
            raise hub.TerminateEventProcessing

    def requestinput(self, event):
        self.inputstring = ""
        self.inputprompt = event.prompt
        self.inputhandler = event.handler
        self.title = event.title
        self.top()
        self.update()

    # window update method
        
    def update(self):
        window.window.update(self)
        
        self.addstr(self.iy, self.ix, " "*(self.iw-1))
        self.move(self.iy, self.ix)
        self.addstr(self.inputprompt, self.colors.description)
        self.addstr(self.inputstring, self.colors.content)


class popupinputwin(inputwin):

    """ input window which appears as a popup at the center of the screen """
    
    def __init__(self, screen, maxh, maxw, channel):
        # calculate size and position 
        h = 3
        w = 40
        y = (maxh-h)/2
        x = (maxw-w)/2

        window.window.__init__(self,
                               screen, h, w, y, x,
                               config.colors.inputwindow,
                               "Input")

        inputwin.__init__(self, screen, maxh, maxw, channel)

    def resize(self, maxh, maxw):
        h = 3
        w = 40
        y = (maxh-h)/2
        x = (maxw-w)/2
        window.window.resize(self, h, w, y, x)

    def requestinput(self, event):
        inputwin.requestinput(self, event)
        if self.inputprompt:
            self.inputprompt = self.inputprompt + " "
            
        sbar = []
        sbar += [("Enter", config.colors.statusbar.key),
                 (": "+_("ok"), config.colors.statusbar.description)]
        sbar += statusbar.separator
        sbar += [("ESC", config.colors.statusbar.key),
                 (": "+_("cancel"), config.colors.statusbar.description)]
        sbar += statusbar.terminate
        hub.notify(events.statusbar_update(0, sbar))

        self.update()


class statusbarinputwin(inputwin):

    """ input window which appears in the statusbar """

    def __init__(self, screen, maxh, maxw, channel):
        # calculate size and position 

        window.window.__init__(self,
                               screen, 1, maxw-1, maxh-1, 0,
                               config.colors.inputwindow)

        inputwin.__init__(self, screen, maxh, maxw, channel)

    def resize(self, maxh, maxw):
        window.window.resize(self, 1, maxw, maxh-1, 0)

    def requestinput(self, event):
        inputwin.requestinput(self, event)
        self.inputprompt = self.title + ": "
        self.update()



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import re
import events, hub

class slist:
    """ Generic list class with selectable items

    List with selectable items, out of which maximally height are
    being displayed at the same time.

    If window corresponding to list has focus, selectionchanged events
    are issued upon change of the currently selected item.
    """

    def __init__(self, win, pagescroll):
        self.win = win
        self.items = []
        self.selected = None   # currently selected item
        self.top = 0           # first displayed item
        self.pagescroll = pagescroll

    # generic list methods

    def __getitem__(self, key):
        return self.items[key]

    def __len__(self):
        return len(self.items)

    def __delitem__(self, index):
        del self.items[index]
        if self.selected is not None:
            if len(self) == 0:
                self.selected = None
            else:
                if index < self.selected:
                    self.selected -= 1
                self.selected = min(self.selected, len(self)-1)
            self._notifyselectionchanged()
        self._updatetop()

    def index(self, item):
        return self.items.index(item)

    def insert(self, index, item):
        self.items.insert(index, item)
        if self.selected is None:
            self.selected = 0
            self._notifyselectionchanged()
        elif self.selected>=index:
            self.selected += 1
            self._notifyselectionchanged()
        self._updatetop()

    def append(self, item):
        self.items.append(item)
        if self.selected is None:
            self.selected = 0
            self._notifyselectionchanged()
        self._updatetop()

    def remove(self, item):
        for i in range(len(self)):
            if self[i] is item:
                del self[i]
                return

    def set(self, items, keepselection=False):
        """set all items in slist trying to keep the current selection if keepselection is set """
        if keepselection:
            oldselecteditem = self.getselected()
            oldselected = self.selected
        try:
            self.items = list(items)
        except:
            self.items = []
        if keepselection and oldselected is not None and self.items:
            # we try to keep the current selection and search for the previously
            # selected item in the new list. We most probably find it around
            # the original position.
            oldselecteditemid = oldselecteditem.getid()
            startsearch = min(max(0, oldselected-1), len(self))
            for i in range(startsearch, len(self)) + range(startsearch):
                if self[i].getid() == oldselecteditemid:
                    self.selected = i
                    break
            else:
                # if this fails (typically because the song has been
                # deleted from the playlist), we take the item at the
                # last selected position, if possible
                if oldselected < len(self):
                    self.selected = oldselected
                elif len(self) > 0:
                    self.selected = len(self)-1
            self._updatetop()
        else:
            if self.items:
                self.selected = 0
            else:
                self.selected = None
            self.top = 0
        self._notifyselectionchanged()

    def sort(self, func=None):
        if func is None:
            self.items.sort()
        else:
            self.items.sort(func)
        self._notifyselectionchanged()

    # helper routines

    def _notifyselectionchanged(self):
        """ helper routine, which issues a selectionchanged event, if window
        corresponding to list has focus """

        if self.win.hasfocus():
            hub.notify(events.selectionchanged(self.getselected()))

    def _updatetop(self):
        "helper routine, which updates self.top"

        if len(self)<=self.win.ih:
            self.top = 0
            return

        if self.selected is not None and self.selected<self.top:
            if self.pagescroll:
                self.top = max(0, self.selected-self.win.ih+1)
            else:
                self.top = self.selected
            return

        if self.selected is not None and self.selected>self.top+self.win.ih-1:
            if self.pagescroll:
                self.top = self.selected
            else:
                self.top = self.selected-self.win.ih+1
            return

    # slist specific methods

    def clear(self):
        "clear list"
        self.items = []
        self.selected = None
        self.top = 0
        self._notifyselectionchanged()

    def insertitem(self, item, cmpfunc):
        "insert item at alphabetically correct position"
        for i in range(len(self)):
            if cmpfunc(self[i], item)>=0:
                self.insert(i, item)
                break
        else:
            self.append(item)

    def getselected(self):
        "return currently selected item"
        if self.selected is not None:
            try:
                return self.items[self.selected]
            except IndexError:
                if self.items:
                    self.selected = len(self.items)-1
                    return self.items[self.selected]
                else:
                    self.selected = None
        return None

    def deleteselected(self):
        "delete currently selected item"
        if self.selected is not None:
            del self[self.selected]

    def selectbynr(self, nr):
        "select nrth entry in list"
        self.selected = nr
        self._notifyselectionchanged()
        self._updatetop()

    def selectbylinenumber(self, nr):
        """select entry by line number in window
        Returns True if selection was valid, otherwise False.
        """
        if len(self) > 0:
            if nr >= 0 and self.top+nr < len(self):
                self.selected = self.top+nr
                self._notifyselectionchanged()
                return True
        return False

    def selectbyid(self, id):
        """select entry by id
        Returns True if selection was valid, otherwise False."""
        if len(self) > 0:
            for i in range(len(self)):
                if self[i].id == id:
                    self.selected = i
                    self._notifyselectionchanged()
                    self._updatetop()
                    return True
        return False

    def selectbysearchstring(self, searchstring):
        """select next entry matching searchstring.
        Returns True if selection was valid, otherwise False."""
        if len(self) > 0:
            searchstring = searchstring.lower()
            if self.selected is None:
                first = 0
            else:
                first = self.selected
            for i in range(first+1, len(self)) + range(first):
                if self[i].getname().lower().find(searchstring)!=-1:
                    self.selected = i
                    self._notifyselectionchanged()
                    self._updatetop()
                    return True
        return False

    def selectbyregexp(self, regexp, includeselected=True):
        """select next entry matching regexp
        Returns True if selection was valid, otherwise False."""
        if len(self) > 0:
            try:
                cregexp = re.compile(regexp, re.IGNORECASE)
            except:
                return
            if self.selected is None:
                first = 0
            else:
                first = self.selected
            for i in range(first + (not includeselected and 1 or 0), len(self)) + range(first):
                if cregexp.search(self[i].getname()):
                    self.selected = i
                    self._notifyselectionchanged()
                    self._updatetop()
                    return True
        return False

    def selectbyletter(self, letter):
        """select next entry beginning with letter
        Returns True if selection was valid, otherwise False."""
        if len(self)>0:
            if self.selected is None:
                first = 0
            else:
                first = self.selected
                letter = letter.lower()
                for i in range(first+1, len(self)) + range(first):
                    if self[i].getname().lower().startswith(letter):
                        self.selected = i
                        self._notifyselectionchanged()
                        self._updatetop()
                        return True
        return False

    def selectrelative(self, dist):
        "change selection relatively by dist"
        if len(self) > 0:
            self.selected += dist
            self.selected = max(self.selected,0)
            self.selected = min(self.selected, len(self)-1)
            self._notifyselectionchanged()
            self._updatetop()

    def selectfirst(self):
        "select first item of list"
        if len(self) > 0:
            self.selected = 0
            self._notifyselectionchanged()
            self._updatetop()

    def selectlast(self):
        "select last item of list"
        if len(self) > 0:
            self.selected = len(self)-1
            self._notifyselectionchanged()
            self._updatetop()

    def selectnext(self):
        "select next item of list"
        self.selectrelative(1)

    def selectprev(self):
        "select previous item of list"
        self.selectrelative(-1)

    def selectnextpage(self):
        "select next page of list"
        self.selectrelative(self.win.ih)

    def selectprevpage(self):
        "select previous page of list"
        if len(self) > 0:
            if self.top-self.win.ih >= 0:
                self.top = self.top-self.win.ih
                self.selected = min(self.top+self.win.ih-1, len(self)-1)
            else:
                self.top = self.selected = 0
            self._notifyselectionchanged()

    def moveitemup(self):
        "move selected item up, if not first"
        if self.selected is not None and self.selected > 0:
            self.items[self.selected-1], self.items[self.selected] = \
            self.items[self.selected], self.items[self.selected-1]
            self.selected -= 1
            self._updatetop()

    def moveitemdown(self):
        "move selected item down, if not last"
        if self.selected is not None and self.selected < len(self)-1:
            self.items[self.selected], self.items[self.selected+1] = \
            self.items[self.selected+1], self.items[self.selected]
            self.selected += 1
            self._updatetop()



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import config
import events, hub, requests
import slist

#
# playlist class, which acts as glue layer between the playlist service
# and the playlist window
#

class playlist(slist.slist):
    def __init__(self, win, playerid):
        slist.slist.__init__(self, win, config.playlistwindow.scrollmode=="page")
        self.playerid = playerid
        self.songdbid = "main"
        items, self.ptime, self.ttime, self.autoplaymode, self.playingitem = \
               hub.request(requests.playlistgetcontents())
        self.set(items)
        self._recenter()

        self.win.channel.subscribe(events.playlistchanged, self.playlistchanged)

    def _recenter(self):
        """ recenter playlist around currently playing (or alternatively last) song """
        for i in range(len(self)):
            if self[i] is self.playingitem or i==len(self)-1:
                oldselected = self.selected
                self.selected = i
                if self.selected != oldselected:
                    self._notifyselectionchanged()
                h2 = self.win.ih/2
                if len(self)-i <= h2:
                    self.top = max(0, len(self)-self.win.ih)
                elif i >= h2:
                    self.top = i-h2
                else:
                    self.top = 0
                self._updatetop()
                break

    def getselectedsong(self):
        """ return song corresponding to currently selected item or None """
        playlistitem = self.getselected()
        if playlistitem:
            return playlistitem.song
        else:
            return None

    # The following three slist.slist methods are delegated to the
    # playlist service. Any resulting changes to the playlist will be
    # performed only later when a playlistchanged event comes back

    def deleteselected(self):
        "delete currently selected item"
        if self.selected is not None:
            hub.notify(events.playlistdeletesong(self.getselected().id))

    def moveitemup(self):
        "move selected item up, if not first"
        if self.selected is not None and self.selected>0:
            hub.notify(events.playlistmovesongup(self.getselected().id))

    def moveitemdown(self):
        "move selected item down, if not last"
        if self.selected is not None and self.selected<len(self)-1:
            hub.notify(events.playlistmovesongdown(self.getselected().id))

    def rateselection(self, rating):
        if self.selected is not None:
            self.getselectedsong().rate(rating)

    def rescanselection(self, force):
        if self.selected is not None:
            song = self.getselectedsong()
            hub.notify(events.autoregisterer_rescansongs(song.songdbid, [song], force))

    def playselected(self):
        item = self.getselected()
        if item:
            hub.notify(events.playlistplaysong(item.id))

    def filelistjumptoselected(self):
        song = self.getselectedsong()
        if song is not None:
            hub.notify(events.filelistjumptosong(song))

    # event handler

    def playlistchanged(self, event):
        self.set(event.items, keepselection=True)
        self.ptime = event.ptime
        self.ttime = event.ttime
        self.autoplaymode = event.autoplaymode
        self.playingitem = event.playingitem

        # recenter window if the playlist window doesn't have the focus.
        if not self.win.hasfocus():
            self._recenter()

        self.win.update()



## -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2003, 2004, 2007 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

class event:
    def __repr__(self):
        return self.__class__.__name__


class dbevent(event):

    """ base class for all database service events """

    def __init__(self, songdbid):
        self.songdbid = songdbid

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.songdbid)


class quit(event):
    """ request end of thread """
    pass


class keypressed(event):
    def __init__(self, key):
        self.key = key

    def __repr__(self):
        return "%r(%d)" % (self.__class__.__name__, self.key)


class mouseevent(event):
    def __init__(self,  y, x, state):
        self.y, self.x, self.state = y, x, state

    def __repr__(self):
        return "%r(%d, %d, %d)" % (self.__class__.__name__,
                                   self.y, self.x, self.state)


class selectionchanged(event):
    def __init__(self, item):
        self.item = item

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.item)


class focuschanged(event):
    pass


class activateplaylist(event):
    pass


class activatefilelist(event):
    pass


class sendeventat(event):

    """ send event at alarmtime and every repeat seconds after (if
    nonzero) or replace the given event"""

    def __init__(self, event, alarmtime,  replace=0):
        self.event = event
        self.alarmtime = alarmtime
        self.repeat = repeat
        self.replace = replace

    def __repr__(self):
        return "%r(%r, %r, %r, %r)" % (self.__class__.__name__, self.event, self.alarmtime, self.repeat, self.replace)


class sendeventin(event):

    """ send event in alartimediff seconds and every repeat seconds
    after (if nonzero), or replace the given event"""

    def __init__(self, event, alarmtimediff, repeat=0, replace=0):
        self.event = event
        self.alarmtimediff = alarmtimediff
        self.repeat = repeat
        self.replace = replace

    def __repr__(self):
        return "%r(%r, %r, %r, %r)" % (self.__class__.__name__, self.event, self.alarmtimediff, self.repeat, self.replace)


class checkpointdb(dbevent):
    """flush memory pool, write checkpoint record to log and flush flog of songdbid"""


class add_song(dbevent):
    """ add song to database """

    def __init__(self, songdbid, song):
        self.songdbid = songdbid
        self.song = song

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.song, self.songdbid)


class update_song(dbevent):
    """ update song in database """

    def __init__(self, songdbid, song):
        self.songdbid = songdbid
        self.song = song

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.song, self.songdbid)


class song_played(dbevent):
    """ register playing of song in database """

    def __init__(self, songdbid, song, date_played):
        self.songdbid = songdbid
        self.song = song
        self.date_played = date_played

    def __repr__(self):
        return "%r(%r, %r)->%r" % (self.__class__.__name__, self.song, self.date_played, self.songdbid)


class song_skipped(dbevent):
    """ register skipping of song in database """

    def __init__(self, songdbid, song):
        self.songdbid = songdbid
        self.song = song

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.song, self.songdbid)


class delete_song(dbevent):
    """ delete song from database """
    def __init__(self, songdbid, song):
        self.songdbid = songdbid
        self.song = song

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.song, self.songdbid)


class add_playlist(dbevent):
    """ add playlist to database """
    def __init__(self, songdbid, name, songs):
        self.songdbid = songdbid
        self.name = name
        self.songs = songs

    def __repr__(self):
        return "%r(%r,%r)->%r" % (self.__class__.__name__, self.name, self.songs, self.songdbid)


class update_playlist(dbevent):
    """ update playlist in database """

    def __init__(self, songdbid, name, songs):
        self.songdbid = songdbid
        self.name = name
        self.songs = songs

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.name, self.songs, self.songdbid)


class delete_playlist(dbevent):
    """ delete playlist from database """

    def __init__(self, songdbid, name):
        self.songdbid = songdbid
        self.name = name

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.name, self.songdbid)




class autoregistersongs(dbevent):
    """ start autoregisterer for database 

    If force is set, the m_time of songs is ignored and they are always rescanned.
    """
    def __init__(self, songdbid, force=False):
        self.songdbid = songdbid
        self.force = force

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.force, self.songdbid)


class autoregisterer_rescansongs(dbevent):
    """ rescan songs in given database 

    If force is set, the m_time of songs is ignored and they are always rescanned
    """
    def __init__(self, songdbid, songs, force=False):
        self.songdbid = songdbid
        self.songs = songs
        self.force = force

    def __repr__(self):
        return "%r(%r, %r)->%r" % (self.__class__.__name__, self.songs, self.force, self.songdbid)


class clearstats(dbevent):
    """ clear playing and added information of all songs (be carefull!) """

    def __init__(self, songdbid):
        self.songdbid = songdbid

    def __repr__(self):
        return "%r->%r" % (self.__class__.__name__, self.songdbid)


class songchanged(event):
    """ song information changed """
    def __init__(self, songdbid, song):
        self.songdbid = songdbid
        self.song = song

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.song, self.songdbid)


class songschanged(event):
    "list of songs in database changed"
    def __init__(self, songdbid):
        self.songdbid = songdbid

    def __repr__(self):
        return "%r->%r" % (self.__class__.__name__, self.songdbid)

class artistschanged(event):
    "list of artists in database changed"
    def __init__(self, songdbid):
        self.songdbid = songdbid

    def __repr__(self):
        return "%r->%r" % (self.__class__.__name__, self.songdbid)


class albumschanged(event):
    "list of albums in database changed"
    def __init__(self, songdbid):
        self.songdbid = songdbid

    def __repr__(self):
        return "%r->%r" % (self.__class__.__name__, self.songdbid)


class tagschanged(event):
    "list of tags in database changed"
    def __init__(self, songdbid):
        self.songdbid = songdbid

    def __repr__(self):
        return "%r->%r" % (self.__class__.__name__, self.songdbid)


class dbplaylistchanged(event):
    def __init__(self, songdbid, playlist):
        self.songdbid = songdbid
        self.playlist = playlist

    def __repr__(self):
        return "%r(%r)->%r" % (self.__class__.__name__, self.playlist, self.songdbid)


class playerevent(event):
    """ event for the player control """

    def __init__(self, playerid):
        self.playerid = playerid

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.playerid)


class playerstart(playerevent):
    """ start player """
    pass


class playerpause(playerevent):
    """ pause player """
    pass


class playertogglepause(playerevent):
    """ pause player, if playing, or start playing, if paused """
    pass


class playernext(playerevent):
    """ play next song on player """
    pass


class playerprevious(playerevent):
    """ play previous song on player"""
    pass


class playerseekrelative(playerevent):
    """ seek relative in song by the given number of seconds """
    
    def __init__(self, playerid, seconds):
        self.playerid = playerid
        self.seconds = seconds

    def __repr__(self):
        return "%r(%f->%r)" % (self.__class__.__name__, self.seconds, self.playerid)

class player_change_volume_relative(playerevent):
    """ change volume of internal player by volume_adj percent"""

    def __init__(self, playerid, volume_adj):
        self.playerid = playerid
        self.volume_adj = volume_adj

class playerplayfaster(playerevent):
    """ increase play speed of song on player"""

    def __init__(self,playerid):
        self.playerid = playerid
        self.speed_adj = 441

class playerplayslower(playerevent):
    """ decrease play speed of song on player"""

    def __init__(self,playerid):
        self.playerid = playerid
        self.speed_adj = 441

class playerspeedreset(playerevent):
    """ Reset play speed of song on player back to its original rate"""

    def __init__(self,playerid):
        self.playerid = playerid

class playerstop(playerevent):
    """ stop player """
    pass


class playerplaysong(playerevent):
    """ play song or playlistitem on player """

    def __init__(self, playerid, playlistitemorsong):
        self.playerid = playerid
        self.playlistitemorsong = playlistitemorsong

    def __repr__(self):
        return "%r(%r->%r)" % (self.__class__.__name__, self.playlistitemorsong, self.playerid)


class playerratecurrentsong(playerevent):
    """ rate song currently being played """
    def __init__(self, playerid, rating):
        playerevent.__init__(self, playerid)
        self.rating = rating

    def __repr__(self):
        return "%r(%r,%d)" % (self.__class__.__name__, self.playerid, self.rating)


class playbackinfochanged(event):
    def __init__(self, playbackinfo):
        self.playbackinfo = playbackinfo

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.playbackinfo)

class player_volume_changed(event):
    """ volume of player has been changed """
    def __init__(self, playerid, volume):
        self.playerid = playerid
        self.volume = volume

    def __repr__(self):
        return "%r(%r,%r)" % (self.__class__.__name__, self.playerid, self.volume)


class statusbar_update(event):
    """ update status bar

    pos = 0: info for currently selected window
    pos = 1: player info
    pos = 2: global info
    """

    def __init__(self, pos, content):
        self.pos = pos
        self.content = content

    def __repr__(self):
        return "%r(%r, %r)" % (self.__class__.__name__, self.pos, self.content)


class statusbar_showmessage(event):
    """ show a message (which automatically disappears after some time in the statusbar"""

    def __init__(self, message):
        self.message = message

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.message)



class requestinput(event):
    def __init__(self, title, prompt, handler):
        self.title = title
        self.prompt = prompt
        self.handler = handler

    def __repr__(self):
        return "%r(%r,%r,%r)" % (self.__class__.__name__,
                              self.title, self.prompt, self.handler)


class playlistevent(event):
    pass


class playlistaddsongs(playlistevent):
    """ add songs to playlist """
    def __init__(self, songs):
        self.songs = songs

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.songs)


class playlistaddsongtop(playlistevent):
    """ add song to top of playlist """
    def __init__(self, song):
        self.song = song

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.song)


class playlistdeletesong(playlistevent):
    def __init__(self, id):
        self.id = id

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.id)


class playlistmovesongup(playlistevent):
    def __init__(self, id):
        self.id = id

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.id)


class playlistmovesongdown(playlistevent):
    def __init__(self, id):
        self.id = id

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.id)


class playlistsave(playlistevent):
    pass


class playlistclear(playlistevent):
    pass


class playlistdeleteplayedsongs(playlistevent):
    pass


class playlistreplay(playlistevent):
    """mark all songs of playlist unplayed again"""
    pass


class playlistshuffle(playlistevent):
    pass


class playlisttoggleautoplaymode(playlistevent):
    pass


class playlistplaysong(playlistevent):
    """ immediately play song in playlist """
    def __init__(self, id):
        self.id = id

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.id)


class playlistchanged(event):
    def __init__(self, items, ptime, ttime, autoplaymode, playingitem):
        self.items = items
        self.ptime = ptime
        self.ttime = ttime
        self.autoplaymode = autoplaymode
        self.playingitem = playingitem

    def __repr__(self):
        return "%r(%r,%r/%r,%r,%r)" % (self.__class__.__name__,
                                       self.items, self.ptime, self.ttime, self.autoplaymode,
                                       self.playingitem)

class filelistjumptosong(event):
    """ jump to specific song in filelist window """
    def __init__(self, song):
        self.song = song

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.song)

class hidewindow(event):
    def __init__(self, window):
        self.window = window

    def __repr__(self):
        return "%r(%r)" % (self.__class__.__name__, self.window)



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import inspect
import os.path
import sys
import threading
import time
import traceback
import codecs

_DEBUG   = 0
_INFO    = 1
_WARNING = 2
_ERROR   = 3

_desc = { _DEBUG:   "D",
          _INFO:    "I",
          _WARNING: "W",
          _ERROR:   "E" }

# minimal level of log messages that should be stored in the buffer
# accesible to the user in the messagewin
LOGLEVEL = _INFO
# LOGLEVEL = _DEBUG

# open file for log output, if necessary
debugfile = None

# length of path prefix (used to obtain module name from path)
pathprefixlen = len(os.path.dirname(__file__))

def initdebugfile(debugfilename):
    """ direct debugging output to debugfilename """
    global debugfile
    if debugfilename:
        debugfile = codecs.open(debugfilename, "w", "utf-8")

# log buffer consisting of tuples (loglevel, time, logmessage)
items = []

# maximal length of log buffer
maxitems = 100

def log(s, level):
    if debugfile:
        try:
            frame = inspect.stack()
            try:
                timestamp = time.strftime("%H:%M:%S", time.localtime())
                threadname = threading.currentThread().getName()
                modulename = frame[2][1][pathprefixlen+1:-3]
                debugfile.write("%s [%s|%s|%s] %s\n" % (_desc[level], timestamp, threadname, modulename, s))
            finally:
                del frame
        except:
            debugfile.write("%s [???] %s\n" % (_desc[level], s))

    if level >= LOGLEVEL:
        items.append((level, time.time(), s))
        if len(items) > maxitems:
            items.pop(0)

def debug(s):
    log(s, _DEBUG)

def info(s):
    log(s, _INFO)

def warning(s):
    log(s, _WARNING)

def error(s):
    log(s, _ERROR)

def debug_traceback():
    debug("Exception caught: %s " % sys.exc_info()[1])
    tblist = traceback.extract_tb(sys.exc_info()[2])
    for s in traceback.format_list(tblist):
        debug(s[:-1])



#!/usr/bin/env python
# -*- coding: ISO-8859-1 -*-
from distutils.core import setup, Extension
import sys; sys.path.append("src")
from version import version

# build extension module which adds transparency support for terminals
# supporting this feature (not necessary for Python 2.4 and above)
# You need the curses header files for building
buildcursext = False

# build extension module which replaces the Python audio output buffer
# by a C version, which should help preventing audio dropouts.
# You need the libao header files for building (but on the other hand,
# you don't need the pyao extension module when you use bufferedao)
buildbufferedaoext = True

# list of supported locales
locales = ["de", "it", "fr", "pl"]

#
# list of packages
#
packages = ["pytone",
            "pytone.services", "pytone.services.players", "pytone.services.songdbs",
            "pytone.plugins", "pytone.plugins.audioscrobbler"]

#
# list of extension modules to be built
#
ext_modules = [Extension("pytone.pcm", sources=["src/pcm/pcm.c"])]

if buildcursext:
    ext_modules.append(Extension("pytone.cursext",
                                 sources=["src/cursext/cursextmodule.c"],
                                 libraries=["curses"]))


if buildbufferedaoext:
    ext_modules.append(Extension("pytone.bufferedao",
                                 sources=["src/bufferedao.c"],
                                 libraries=["ao"]))
#
# list of data files to be installed
#
mo_files = ["locale/%s/LC_MESSAGES/PyTone.mo" % locale for locale in locales]
data_files=[('share/locale/de/LC_MESSAGES', mo_files)]

#
# list of scripts to be installed
#
# Note that we (ab-)use distutils scripts option to install our wrapper
# files (hopefully) at the correct location.
scripts=['pytone', 'pytonectl']

#
# additional package metadata
#
classifiers = ["Development Status :: 5 - Production/Stable",
               "Environment :: Console :: Curses",
               "Intended Audience :: End Users/Desktop",
               "License :: OSI Approved :: GNU General Public License (GPL)",
               "Programming Language :: Python",
               "Topic :: Multimedia :: Sound/Audio :: Players"]

if sys.version_info >= (2, 3):
    addargs = {"classifiers": classifiers}
else:
    addargs = {}

setup(name="PyTone",
      version=version,
      description="Powerful music jukebox with a curses based GUI.",
      author="Jörg Lehmann",
      author_email="joerg@luga.de",
      url="http://www.luga.de/pytone/",
      license="GPL",
      package_dir={"pytone": "src"},
      packages=packages,
      ext_modules=ext_modules,
      data_files=data_files,
      scripts=scripts,
      **addargs)



### OSD title plugin for pytone
### """""""""""""""""""""""""""
### This plugin displays on-screen song titles whenever pytone changes to
### a new song. You can install this plugin by simply copying it to your
### pytone system-wide plugin directory (eg. /usr/share/pytone/plugins
### or ~/pytone/plugins/). And enable it by adding it to the plugins list
### in the general section, like:
###
###     [general]
###     plugins = osdtitle
###
### You can customise the title by adding the following to /etc/pytonerc
### or ~/.pytone/pytonerc
###
###     [plugin.osdtitle]
###     songformat = %(artist)s - %(title)s (%(length)s)
###     font = -adobe-helvetica-bold-r-*-*-*-240-*-*-*-*-*-*
###     color = green
###     position = top
###     offset = 60
###     align = center
###     shadow = 2
###     lines = 1
###
### See the osd_cat(1) manpage for more information about these options.
###
### You can find the latest release at:
###
###     http://dag.wieers.com/home-made/pytone/
###
### If you have improvements or changes to this plugin, please
### send them to the pytone mailinglist and include me as well:
###
###     pytone-users@luga.de, Dag Wieers <dag@wieers.com>

import events, encoding, plugin, config, os, re

class config(config.configsection):
    songformat = config.configstring('%(artist)s - %(title)s (%(length)s)')
    font = config.configstring('-adobe-helvetica-bold-r-*-*-*-240-*-*-*-*-*-*')
    color = config.configstring('green')
    position = config.configstring('top')
    offset = config.configstring('60')
    align = config.configstring('center')
    shadow = config.configstring('2')
    lines = config.configstring('1')

class plugin(plugin.plugin):
    def init(self):
        self.channel.subscribe(events.playbackinfochanged, self.playbackinfochanged)
        self.command = 'osd_cat -p %(position)s -c %(color)s -o %(offset)s -A %(align)s -s %(shadow)s -l %(lines)s -f %(font)s -' % self.config
        self.previoussong = ''

    def playbackinfochanged(self, event):
        if event.playbackinfo.song and event.playbackinfo.song != self.previoussong:
            song = encoding.encode(event.playbackinfo.song.format(self.config.songformat, safe=True))
            os.system('echo "%s" | %s &' % (song, self.command))
            self.previoussong = event.playbackinfo.song

# vim:ts=4:sw=4



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import service
import config

class plugin:

    def __init__(self, channel, config, mainscreen):
        """ create a plugin instance """
        self.channel = channel
        self.config = config
        self.mainscreen = mainscreen

    def start(self):
        self.init()

    def init(self):
        """ initialize plugin after it has been configured """


class threadedplugin(service.service, plugin):

    def __init__(self, channel, config, mainscreen):
        service.service.__init__(self, self.__class__.__name__)
        # as independent thread, we have to use our own channel which has
        # been created by the service constructor
        plugin.__init__(self, self.channel, config, mainscreen)

    def run(self):
        self.init()
        service.service.run(self)



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2003, 2007 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import ConfigParser, copy, curses, sys, getopt, exceptions, os.path, types, re, types
import log, encoding, version


class ConfigError(Exception):
    pass


class configsection:

    """section of the config file"""

    # we store our configuration items in a separate dictionary _configitems
    # after instantiation, we only allow reading and writing to this
    # dictionary (via the corresponding configitems get and set method)
    def __getattr__(self, name):
        try:
            return self._configitems[name].get()
        except KeyError:
            raise AttributeError
            try:
                return self._configsections[name]
            except KeyError:
                raise AttributeError

    def __setattr__(self, name, value):
        self._configitems[name].set(value)

    def __delattr__(self, name):
        del self._configitems[name]

    def __getitem__(self, name):
        try:
            return self._configitems[name].get()
        except KeyError:
            try:
                return self._configsections[name]
            except KeyError:
                raise IndexError

    def asdict(self):
        d = {}
        for n, v in self._configitems.items():
            d[n] = v
        return d

    def getsubsections(self):
        return self._configsections.keys()

#
# define different types of configuration variables
#

class configitem:
    def __init__(self, default):
        self._check(default)
        self.default = default
        self.value = None
        # cached effective output value
        self._cachedoutput = None

    def _check(self, s):
        """ check whether string conforms with expected format
        If not, this method should raise a ConfigError exception.
        """
        pass

    def _convert(self, s):
        """ convert from string to item value """
        return s

    def set(self, s):
        self._check(s)
        self.value = s
        self._cachedoutput = None

    def get(self):
        if self._cachedoutput is None:
            if self.value is None:
                self._cachedoutput = self._convert(self.default)
            else:
                self._cachedoutput = self._convert(self.value)
        return self._cachedoutput

class configstring(configitem):
    pass


class configint(configitem):
    def _check(self, s):
        try:
            int(s)
        except:
            raise ConfigError("Expecting float, got '%s'" % s)

    def _convert(self, s):
        return int(s)


class configfloat(configitem):
    def _check(self, s):
        try:
            float(s)
        except:
            raise ConfigError("Expecting float, got '%s'" % s)

    def _convert(self, s):
        return float(s)


class configlist(configitem):
    def _check(self, s):
        try:
            items = s.split()
        except:
            raise ConfigError("Expecting list, got '%s'" % s)

    def _convert(self, s):
        return s.split()


class configcolor(configitem):

    # dict of available curses colors
    #
    # The entry for "default" is set to -1 (in mainscreen.py) if the
    # curses.use_default_colors() call succeeds. It then represents a
    # (possibly transpert) default color.

    _colors = { "white": curses.COLOR_WHITE,
                "black": curses.COLOR_BLACK,
                "green": curses.COLOR_GREEN,
                "magenta": curses.COLOR_MAGENTA,
                "blue": curses.COLOR_BLUE,
                "cyan": curses.COLOR_CYAN,
                "yellow": curses.COLOR_YELLOW,
                "red": curses.COLOR_RED,
                "default" : 0 }

    _mono = { "none": curses.A_NORMAL,
              "bold": curses.A_BOLD,
              "underline": curses.A_UNDERLINE,
              "reverse": curses.A_REVERSE,
              "standout": curses.A_STANDOUT }

    _defaultbg = "default"

    # disable color support by default. Reenable it in mainscreen.py
    # if supported by terminal
    _colorenabled = 0

    _colorpairs = []

    # two helper methods which parse a color or a mono definition
    # and return the rest of the line

    def _parsecolor(self, cdef):
        if cdef[1].startswith("bright"):
            fg = self._colors[cdef[1][6:]]
            bright = 1
        else:
            fg = self._colors[cdef[1]]
            bright = 0
        if len(cdef)>2 and cdef[2]!="mono":
            bg = self._colors[cdef[2]]
            return ((fg, bg, bright), cdef[3:])
        else:
            bg = self._colors[self._defaultbg]
            return ((fg, bg, bright), cdef[2:])

    def _parsemono(self, cdef):
        attr = cdef[1]
        return (self._mono[attr], cdef[2:])

    def _parsecolormono(self, cdef):
        # parse combined color and mono definition
        fg = bg = bright = attr = None
        if cdef[0] == "color":
            (fg, bg, bright), cdef = self._parsecolor(cdef)
            if cdef:
                attr, cdef = self._parsemono(cdef)
            else:
                attr = self._mono["none"]
        elif cdef[0] == "mono":
            attr, cdef = self._parsemono(cdef)
            if cdef:
                (fg, bg, bright), cdef = self._parsecolor(cdef)
        if cdef:
            raise ConfigError("color definition too long")
        return fg, bg, bright, attr

    def _check(self, s):
        try:
            self._parsecolormono(s.split())
        except:
            raise ConfigError("wrong color definition '%s'" %s )

    def _convert(self, s):
        fg, bg, bright, attr = self._parsecolormono(s.split())
        if fg is not None and self._colorenabled or attr is None:
            try:
                colorindex = self._colorpairs.index((fg, bg))+1
            except ValueError:
                self._colorpairs.append((fg, bg))
                colorindex = len(self._colorpairs)
                curses.init_pair(colorindex, fg, bg)
            color = curses.color_pair(colorindex)
            if bright:
                color |= curses.A_BOLD
            return color
        else:
            return attr

class configkeys(configitem):

    def _check(self, s):
        if not s:
           return
        for key in s.split(" "):
            keyorig = key
            if key[:5].lower()=="ctrl-":
                key = key[5:].upper()
            elif key[:4].lower()=="alt-":
                key = key[4:]
            if key=="KEY_SPACE":
               pass
            elif key.startswith("KEY_") and key[4:].isalnum():
                try:
                    eval("curses.%s" % key)
                except:
                    raise ConfigError("wrong key specification '%s'" % keyorig)
            elif key.startswith("\\") and len(key)==2 and key[1] in ("n", "r", "t"):
                pass
            elif len(key)!=1:
                raise ConfigError("wrong key specification '%s'" % keyorig)

    def _convert(self, s):
        keys = []
        for key in s.split(" "):
            modifier = 0
            if key[:5].lower()=="ctrl-":
                key = key[5:].upper()
                modifier = -64
            elif key[:4].lower()=="alt-":
                key = key[4:]
                modifier = 1024
            if key=="KEY_SPACE":
                keyvalue = 32
            elif key.startswith("KEY_") and key[4:].isalnum():
                keyvalue = eval("curses.%s" % key)
            elif key.startswith("\\") and len(key)==2:
                keyvalue = ord({"n": "\n", "r": "\r", "t": "\t"}[key[1]])
            elif len(key)==1:
                keyvalue = ord(key)
            keys.append(keyvalue+modifier)
        return keys


class configboolean(configitem):
    def _check(self, s):
        if s not in ("0", "1", "on", "off", "true", "false"):
            raise ConfigError("Excepting boolean, got '%s'" % s)

    def _convert(self, s):
        return s in ("1", "on", "true")


class configalternatives(configitem):
    def __init__(self, default, alternatives):
        self.alternatives = alternatives
        configitem.__init__(self, default)

    def _check(self, s):
        if s not in self.alternatives:
            raise ConfigError("Expecting one of %s, got %s" % (str(self.alternatives), s))


class configpath(configitem):
    def _convert(self, s):
        if s != "":
            return os.path.normpath(os.path.expanduser(s))
        else:
            return ""


class confignetworklocation(configitem):

    def _parselocation(self, s):
        if ":" in s:
            # address:port
            name, port = s.split(":")
            port = int(port)
            return name, port
        else:
            return os.path.expanduser(s)

    def _check(self, s):
        try:
            self._parselocation(s)
        except:
            raise ConfigError("Excepting address:port or filename, got '%s'" % s)            

    def _convert(self, s):
        return self._parselocation(s)
 

class configre(configitem):
    def _convert(self, s):
        return re.compile(s)




BORDER_TOP = 1
BORDER_BOTTOM = 2
BORDER_LEFT = 4
BORDER_RIGHT = 8
BORDER_COMPACT = 16
BORDER_ULTRACOMPACT = 32

class configborder(configitem):
    def _check(self, s):
        if s == "all" or s == "compact" or s == "off" or s == "ultracompact":
            return
        for b in s.split():
            if b not in ("top", "bottom", "left", "right", "compact"):
                raise ConfigError("Expecting one of 'top', 'bottom', 'left', or 'right', got '%s'" % b)
    def _convert(self, s):
        result = 0
        if s == "all":
            return BORDER_TOP | BORDER_BOTTOM | BORDER_LEFT | BORDER_RIGHT
        if s == "compact":
            return BORDER_COMPACT
        if s == "off":
            return 0
        if s == "ultracompact":
            return BORDER_ULTRACOMPACT
        for b in s.split():
            if b == "top":
                result |= BORDER_TOP
            elif b == "bottom":
                result |= BORDER_BOTTOM
            elif b == "left":
                result |= BORDER_LEFT
            elif b == "right":
                result |= BORDER_RIGHT
        return result

##############################################################################
# configuration tree
##############################################################################

class general(configsection):
    logfile = configpath("~/.pytone/pytone.log")
    songchangecommand = configstring("")
    playerinfofile = configpath("~/.pytone/playerinfo")
    dumpfile = configpath("~/.pytone/pytone.dump")
    debugfile = configpath("")
    randominsertlength = configfloat("3600")
    colorsupport = configalternatives("auto", ["auto", "on", "off"])
    mousesupport = configboolean("on")
    layout = configalternatives("twocolumn", ["onecolumn", "twocolumn"])
    throttleoutput = configint("0")
    autoplaymode = configalternatives("off", ["off", "repeat", "random"])
    plugins = configlist("")

class database(configsection):
    requestcachesize = configint("50000")
    class __template__(configsection):
        type = configalternatives("local", ["local", "remote"])
        dbfile = configpath("~/.pytone/main.db")
        cachesize = configint("1000")
        musicbasedir = configpath("")
        tracknrandtitlere = configre(r"^\[?(\d+)\]? ?[- ] ?(.*)\.(mp3|ogg)$")
        postprocessors = configlist("capitalize strip_leading_article add_decade_tag")
        autoregisterer = configboolean("on")
        playingstatslength = configint("100")
        networklocation = confignetworklocation("localhost:1972")

class tag(configsection):
    class __template__(configsection):
        name = configstring("")
        key = configkeys("")


class mixer(configsection):
    type = configalternatives("external", ["external", "internal", "off"])
    device = configpath("/dev/mixer")
    channel = configstring("SOUND_MIXER_PCM")
    stepsize = configint("5")

class network(configsection):
    socketfile = configstring("~/.pytone/pytonectl")
    enableserver = configboolean("false")
    bind = configstring("")
    port = configint("1972")

class player(configsection):
    class main(configsection):
        type = configalternatives("internal", ["internal", "xmms", "mpg123", "remote", "off"])
        autoplay = configboolean("true")

        # only for internal player
        driver = configalternatives("oss", ["alsa", "alsa09", "alsa05", "arts", "esd", "oss", "sun", "macosx", "macosxau", "pulse"])
        device = configstring("/dev/dsp")
        bufsize = configint(100)
        crossfading = configboolean("true")
        crossfadingstart = configfloat(5)
        crossfadingduration = configfloat(6)
        aooptions = configstring("")

        # only for xmms player
        session = configint("0")
        noqueue = configboolean("false")

        # only for mpg123 player
        cmdline = configstring("/usr/bin/mpg321 --skip-printing-frames=5 -a /dev/dsp")

        # only for remote player
        networklocation = confignetworklocation("localhost:1972")

    class secondary(configsection):
        type = configalternatives("off", ["internal", "xmms", "mpg123", "off"])
        autoplay = configboolean("true")

        # only for internal player
        driver = configalternatives("oss", ["alsa", "alsa09", "alsa05", "arts", "esd", "oss", "sun", "macosx", "pulse"])
        device = configstring("/dev/dsp1")
        bufsize = configint(100)
        crossfading = configboolean("true")
        crossfadingstart = configfloat(5)
        crossfadingduration = configfloat(6)
        aooptions = configstring("")

        # only for xmms player
        session = configint("0")
        noqueue = configboolean("false")

        # only for mpg123 player
        cmdline = configstring("/usr/bin/mpg321 --skip-printing-frames=5 -a /dev/dsp1")

class filelistwindow(configsection):
    border = configborder("all")
    scrollbar = configboolean("true")
    scrollmode = configalternatives("page", ["page", "line"])
    virtualdirectoriesattop = configboolean("false")
    skipsinglealbums = configboolean("true")


class playerwindow(configsection):
    border = configborder("all")
    songformat = configstring("%(artist)s - %(title)s")


class iteminfowindow(configsection):
    border = configborder("all")


class playlistwindow(configsection):
    border = configborder("all")
    scrollbar = configboolean("true")
    scrollmode = configalternatives("page", ["page", "line"])
    songformat = configstring("%(artist)s - %(title)s")


class mixerwindow(configsection):
    type = configalternatives("popup", ["popup", "statusbar"])
    autoclosetime = configfloat("5")


class helpwindow(configsection):
    autoclosetime = configfloat("0")


class logwindow(configsection):
    autoclosetime = configfloat("10")


class statswindow(configsection):
    autoclosetime = configfloat("10")


class iteminfolongwindow(configsection):
    autoclosetime = configfloat("30")


class lyricswindow(configsection):
    autoclosetime = configfloat("0")


class inputwindow(configsection):
    type = configalternatives("popup", ["popup", "statusbar"])


class colors(configsection):
    class filelistwindow(configsection):
        title = configcolor("color brightgreen mono bold")
        activetitle = configcolor("color brightgreen mono bold")
        background = configcolor("color white")
        selected_song = configcolor("color white red mono reverse")
        artist_album = configcolor("color brightblue mono bold")
        directory = configcolor("color brightcyan mono bold")
        border = configcolor("color green")
        activeborder = configcolor("color brightgreen mono bold")
        scrollbar = configcolor("color green")
        scrollbarhigh = configcolor("color brightgreen mono bold")
        scrollbararrow = configcolor("color brightgreen mono bold")
        song = configcolor("color white")
        selected_directory = configcolor("color brightcyan red mono reverse")
        selected_artist_album = configcolor("color brightblue red mono reverse")

    class playlistwindow(configsection):
        title = configcolor("color brightgreen mono bold")
        activetitle = configcolor("color brightgreen mono bold")
        background = configcolor("color white")
        unplayedsong = configcolor("color brightwhite mono bold")
        selected_unplayedsong = configcolor("color brightwhite red mono reverse")
        playedsong = configcolor("color white")
        selected_playedsong = configcolor("color white red mono reverse")
        playingsong = configcolor("color yellow mono underline")
        selected_playingsong = configcolor("color yellow red mono reverse")
        border = configcolor("color green")
        activeborder = configcolor("color brightgreen mono bold")
        scrollbar = configcolor("color green")
        scrollbarhigh = configcolor("color brightgreen mono bold")
        scrollbararrow = configcolor("color brightgreen mono bold")

    class playerwindow(configsection):
        title = configcolor("color brightgreen mono bold")
        content = configcolor("color white")
        background = configcolor("color white")
        description = configcolor("color brightcyan mono bold")
        activeborder = configcolor("color brightgreen mono bold")
        progressbar = configcolor("color cyan cyan")
        border = configcolor("color green")
        progressbarhigh = configcolor("color red red mono bold")

    class iteminfowindow(configsection):
        title = configcolor("color brightgreen mono bold")
        content = configcolor("color white")
        background = configcolor("color white")
        description = configcolor("color brightcyan mono bold")
        activeborder = configcolor("color brightgreen mono bold")
        border = configcolor("color green")

    class iteminfolongwindow(configsection):
        title = configcolor("color brightgreen mono bold")
        content = configcolor("color white")
        background = configcolor("color white")
        description = configcolor("color brightcyan mono bold")
        activeborder = configcolor("color brightgreen mono bold")
        border = configcolor("color green")

    class lyricswindow(configsection):
        title = configcolor("color brightgreen mono bold")
        content = configcolor("color white")
        background = configcolor("color white")
        activeborder = configcolor("color brightgreen mono bold")
        border = configcolor("color green")

    class inputwindow(configsection):
        title = configcolor("color brightgreen mono bold")
        content = configcolor("color white")
        background = configcolor("color white")
        description = configcolor("color brightcyan mono bold")
        activeborder = configcolor("color brightgreen mono bold")
        border = configcolor("color green")

    class mixerwindow(configsection):
        title = configcolor("color brightgreen mono bold")
        content = configcolor("color white")
        bar = configcolor("color cyan cyan")
        description = configcolor("color brightcyan mono bold")
        border = configcolor("color green")
        background = configcolor("color white")
        activeborder = configcolor("color brightgreen mono bold")
        barhigh = configcolor("color red red mono bold")

    class helpwindow(configsection):
        title = configcolor("color brightgreen mono bold")
        background = configcolor("color white")
        key = configcolor("color brightcyan mono bold")
        description = configcolor("color white")
        activeborder = configcolor("color brightgreen mono bold")
        border = configcolor("color green")

    class logwindow(configsection):
        title = configcolor("color brightgreen mono bold")
        background = configcolor("color white")
        time = configcolor("color brightcyan mono bold")
        debug = configcolor("color white")
        info = configcolor("color white")
        warning = configcolor("color cyan")
        error = configcolor("color red mono bold")
        activeborder = configcolor("color brightgreen mono bold")
        border = configcolor("color green")

    class statswindow(configsection):
        title = configcolor("color brightgreen mono bold")
        content = configcolor("color white")
        background = configcolor("color white")
        description = configcolor("color brightcyan mono bold")
        activeborder = configcolor("color brightgreen mono bold")
        border = configcolor("color green")

    class statusbar(configsection):
        key = configcolor("color brightcyan mono bold")
        background = configcolor("color white")
        description = configcolor("color white")


class keybindings(configsection):
    class general(configsection):
        refresh = configkeys("ctrl-l")
        exit = configkeys("ctrl-x")
        playerstart = configkeys("p P")
        playerpause = configkeys("p P")
        playernextsong = configkeys("n N")
        playerprevioussong = configkeys("b B")
        playerforward = configkeys(">")
        playerrewind = configkeys("<")
        playerstop = configkeys("S")
        playlistdeleteplayedsongs = configkeys("KEY_BACKSPACE")
        playlistclear = configkeys("ctrl-d")
        playlistsave = configkeys("ctrl-w")
        playlistreplay = configkeys("ctrl-u")
        playlisttoggleautoplaymode = configkeys("ctrl-t")
        togglelayout = configkeys("KEY_F10")
        showhelp = configkeys("?")
        showlog = configkeys("!")
        showstats = configkeys("%")
        showiteminfolong = configkeys("=")
        showlyrics = configkeys("L")
        toggleiteminfowindow = configkeys("ctrl-v")
        volumeup = configkeys(")")
        volumedown = configkeys("(")
        playerplayfaster = configkeys("}")
        playerplayslower = configkeys("{")
        playerspeedreset = configkeys("~")
        playerratecurrentsong1 = configkeys("alt-1")
        playerratecurrentsong2 = configkeys("alt-2")
        playerratecurrentsong3 = configkeys("alt-3")
        playerratecurrentsong4 = configkeys("alt-4")
        playerratecurrentsong5 = configkeys("alt-5")

    class filelistwindow(configsection):
        selectnext = configkeys("KEY_DOWN j")
        selectprev = configkeys("KEY_UP k")
        selectnextpage = configkeys("ctrl-n KEY_NPAGE")
        selectprevpage = configkeys("ctrl-p KEY_PPAGE")
        selectfirst = configkeys("ctrl-a KEY_HOME")
        selectlast = configkeys("ctrl-e KEY_END")
        dirdown = configkeys("KEY_RIGHT KEY_SPACE \n KEY_ENTER l")
        dirup = configkeys("KEY_LEFT h")
        addsongtoplaylist = configkeys("KEY_SPACE \n KEY_ENTER KEY_RIGHT")
        adddirtoplaylist = configkeys("i I KEY_IC alt-KEY_RIGHT")
        playselectedsong = configkeys("alt-\n alt-KEY_ENTER")
        activateplaylist = configkeys("\t")
        insertrandomlist = configkeys("r R")
        rescan = configkeys("u U")
        toggledelete = configkeys("D")
        search = configkeys("/ ctrl-s")
        repeatsearch = configkeys("ctrl-g")
        focus = configkeys("f")

    class playlistwindow(configsection):
        selectnext = configkeys("KEY_DOWN j")
        selectprev = configkeys("KEY_UP k")
        selectnextpage = configkeys("ctrl-n KEY_NPAGE")
        selectprevpage = configkeys("ctrl-p KEY_PPAGE")
        selectfirst = configkeys("ctrl-a KEY_HOME")
        selectlast = configkeys("ctrl-e KEY_END")
        moveitemup = configkeys("+")
        moveitemdown = configkeys("-")
        deleteitem = configkeys("d D KEY_DC")
        activatefilelist = configkeys("\t KEY_LEFT h")
        playselectedsong = configkeys("alt-\n alt-KEY_ENTER")
        rescan = configkeys("u U")
        shuffle = configkeys("r R")
        filelistjumptoselectedsong = configkeys("KEY_RIGHT l")

#
# register known configuration sections
#

sections = ['mixerwindow', 'helpwindow', 'filelistwindow', 'database', 'tag', 'iteminfowindow',
            'logwindow', 'statswindow', 'iteminfolongwindow', 'lyricswindow', 'mixer', 'colors', 'playerwindow', 'playlistwindow',
            'general', 'inputwindow', 'network', 'player', 'keybindings']

##############################################################################
# end configuration tree
##############################################################################

# options which can be overwritten via the command line

userconfigfile = os.path.expanduser("~/.pytone/pytonerc")
forcedatabaserebuild = False
forcedebugfile = None

#
# helper functions
#

# configparser used for the config files
configparser = None

def setupconfigparser():
    """ initialize ConfigParser.RawConfigParser for the standard configuration files """
    global configparser
    cflist = ["/etc/pytonerc", userconfigfile]
    s = ", ".join(map(os.path.realpath, cflist))
    log.info("Using configuration from file(s) %s" % s)
    configparser = ConfigParser.RawConfigParser()
    configparser.read(cflist)


def readconfigsection(section, clsection):
    """ fill configsection subclass clsection with entries stored under the name section in configfile and
    return a corresponding clsection instance"""
    try:
        for option in configparser.options(section):
            if not clsection.__dict__.has_key(option):
                raise ConfigError("Unkown configuration option '%s' in section '%s'" %
                                  (option, section))
            value = configparser.get(section, option)
            clsection.__dict__[option].set(value)
    except ConfigParser.NoSectionError:
        pass


def finishconfigsection(clsection):
    """ finish a configsection subclass clsection """
    # move configuration items from class __dict__ into _configitems and store
    # config subsections
    clsection._configitems = {}
    clsection._configsections = {}
    for n, v in clsection.__dict__.items():
        if isinstance(v, configitem):
            clsection._configitems[n] = v
            del clsection.__dict__[n]
        elif type(v) == types.ClassType and issubclass(v, configsection) and n != "__template__":
            finishconfigsection(v)
            # instantiate class to make __getattr__ and __setattr__ work and add the instance to
            # a _configsections dictionary
            clsection.__dict__[n] = clsection._configsections[n] = v()


def processstandardconfig():
    for section in configparser.sections():
        if not section.startswith("plugin."):
            if "." not in section:
                # not a subsection
                if section not in sections:
                    raise ConfigError("Unkown configuration section '%s'" % section)
            else:
                # check for valid subsection
                # first check the form (section.subsection) of the config section
                try:
                    mainsection, subsection = section.split(".")
                except:
                    raise ConfigError("Unkown configuration section '%s'" % section)
                if mainsection not in sections or not subsection.isalnum():
                    raise ConfigError("Unkown configuration section '%s'" % section)
                # get corresponding config class
                try:
                    clsection = eval(section)
                    if not issubclass(clsection, configsection):
                        raise ConfigError("Unkown configuration section '%s'" % section)
                except AttributeError:
                    # if it does not exist, use a __template__ class in
                    # mainsection if possible
                    try:
                        # We copy the __template__ class by explicitely
                        # creating a new class with a deeply copied class
                        # dictionary (maybe there is an easier solution)
                        # Any
                        templateclassdict = eval("%s.__template__.__dict__" % mainsection)
                        newclass = types.ClassType(mainsection+subsection, (configsection,), copy.deepcopy(templateclassdict))
                        exec("%s.%s = newclass" % (mainsection, subsection))
                    except AttributeError:
                        raise ConfigError("Unkown configuration section '%s'" % section)

            readconfigsection(section, eval(section))


def finishconfig():
    """ prepare all configuration sections for later use and read command line arguments """

    for section in sections:
        finishconfigsection(eval(section))

        # convert the configsection class into an instance of the same class
        # to make __getattr__ and __setattr__ work
        exec("%s = %s()" % (section, section)) in globals(), globals()

    # apply command line options
    if forcedatabaserebuild:
        for databasename in database.getsubsections():
            exec("database.%s.autoregisterer = 'on'" % databasename) in globals(), globals()

    if forcedebugfile:
        general.debugfile = forcedebugfile


def gendefault():
    cp = ConfigParser()
    cp.read("/home/ringo/PyTone/config")
    for section in cp.sections():
        print "class %s(configsection):" % section
        for option in cp.options(section):
            default = cp.get(section, option)

            if option.startswith("color_"):
                itemname = "configcolor"
            elif default in ("on", "off", "true", "false"):
                itemname = "configboolean"
            else:
                try:
                    float(default)
                    itemname = "configfloat"
                except:
                    itemname = "configstring"

            print '    %s = %s("%s")' % (option, itemname, default)
        print

    print "sections =", cp.sections()

#
# parse command line options
#

def usage():
    print "PyTone %s" % version.version
    print "Copyright %s" % encoding.encode(version.copyright)
    print "usage: pytone.py [options]"
    print "-h, --help: show this help"
    print "-c, --config <filename>: read config from filename"
    print "-d, --debug <filename>: enable debugging output (into filename)"
    print "-r, --rebuild: rebuild all databases"


def processcommandline():
    # we pass the information on the command line options via
    # global variables
    global userconfigfile
    global forcedatabaserebuild
    global forcedebugfile
    try:
        # keep rest of arguments for other use
        global args
        opts, args = getopt.getopt(sys.argv[1:],
                                   "hc:d:r",
                                   ["help", "config=", "debug=", "rebuild"])
    except getopt.GetoptError:
        usage()
        sys.exit(2)

    for o, a in opts:
        if o in ("-h", "--help"):
            usage()
            sys.exit()
        if o in ("-d", "--debug"):
            forcedebugfile = a
        if o in ("-c", "--config"):
            userconfigfile = a
        if o in ("-r", "--rebuild"):
            forcedatabaserebuild = True


def checkoptions():
    #if network.enableserver and network.server:
    #    usage()
    #    print "Error: cannot run both as server and as client"
    #    sys.exit(2)


    import metadata
    # check database options
    dbfiles = []

    if not database.getsubsections():
        print "Please define at least one song database in your configuration file."
        sys.exit(2)

    for databasename in database.getsubsections():
        songdb = database[databasename]

        if songdb.type == "local" and songdb.musicbasedir == "":
            print ( "Please set musicbasedir in the [database.%s] section of the config file ~/.pytone/pytonerc\n"
                    "to the location of your MP3/Ogg Vorbis files." % databasename )
            sys.exit(2)

        if songdb.dbfile != "":
            if songdb.dbfile in dbfiles:
                print "dbfile '%s' of database '%s' already in use." % (songdb.dbfile, databasename)
                sys.exit(2)
            dbfiles.append(songdb.dbfile)

    # check ao options
    for aooption in player.main.aooptions.split() + player.secondary.aooptions.split():
        if aooption.count("=")!=1:
            raise RuntimeError("invalid format for alsa option '%s'" % aooption)


def processconfig():
    setupconfigparser()
    processstandardconfig()
    finishconfig()
    checkoptions()



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import time
import events, hub
import service

class timer(service.service):

    """ service which sends events at specified times """

    def __init__(self):
        """ constructs timer, which sends its events through eventhub """
        service.service.__init__(self, "timer")
        
        # each element in alarms is a tuple (alarmtime, event, repeat)
        self.alarms = []

        self.channel.subscribe(events.sendeventat, self.sendeventat)        
        self.channel.subscribe(events.sendeventin, self.sendeventin)

    def work(self):
        # TODO we could look for the next event and set the
        # timeout accordingly
        self.channel.process(block=True, timeout=0.5)
        acttime = time.time()
        for alarmtime, event, repeat in self.alarms:
            if alarmtime <= acttime:
                hub.notify(event)
                self.alarms.remove((alarmtime, event, repeat))
                if repeat:
                    self.alarms.append((alarmtime+repeat, event, repeat))

    def _sendeventat(self, event, alarmtime, repeat, replace):
        if replace:
            for i in range(len(self.alarms)):
                aalarmtime, aevent, arepeat = self.alarms[i]
                if aevent is event and arepeat==repeat:
                    self.alarms[i] = (alarmtime, event, repeat)
                    return

        self.alarms.append((alarmtime, event, repeat))

    # event handlers

    def sendeventat(self, event, alarmtime, repeat=False, replace=False):
        self._sendeventat(event.event, event.alarmtime, event.repeat, event.replace)
        
    def sendeventin(self, event):
        acttime = time.time()
        self._sendeventat(event.event, acttime+event.alarmtimediff, event.repeat, event.replace)



# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2002, 2007 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import curses

import config
import events, hub
import item
import filelist
import statusbar
import window
import encoding


class filelistwin(window.window):
    def __init__(self, screen, layout, channel, songdbids):
        self.channel = channel
        self.keybindings = config.keybindings.filelistwindow
        self.songdbids = songdbids

        # last search string
        self.searchstring = None
        # list of selections during incremental search
        self.searchpositions = []

        # last song added to playlist
        self.lastadded = None

        h, w, y, x, border = layout

        window.window.__init__(self,
                               screen, h, w, y, x,
                               config.colors.filelistwindow,
                               "MP3s",
                               border, config.filelistwindow.scrollbar)

        self.items = filelist.filelist(self, self.songdbids)

        self.channel.subscribe(events.keypressed, self.keypressed)
        self.channel.subscribe(events.mouseevent, self.mouseevent)
        self.channel.subscribe(events.focuschanged, self.focuschanged)

    def sendmessage(self, message):
        hub.notify(events.statusbar_showmessage(message))
        # allow message to be processed
        self.channel.process()

    def updatestatusbar(self):
        sbar = []
        if len(self.items.shistory)>0:
            sbar += statusbar.generatedescription("filelistwindow", "dirup")
            sbar += statusbar.separator
        if self.items.isdirselected():
            sbar += statusbar.generatedescription("filelistwindow", "dirdown")
            sbar += statusbar.separator
            sbar += statusbar.generatedescription("filelistwindow", "adddirtoplaylist")
            sbar += statusbar.separator
        elif self.items.issongselected():
            sbar += statusbar.generatedescription("filelistwindow", "addsongtoplaylist")
            sbar += statusbar.separator

        sbar += statusbar.generatedescription("filelistwindow", "activateplaylist")
        hub.notify(events.statusbar_update(0, sbar))

    def updatescrollbar(self):
        self.drawscrollbar(self.items.top, len(self.items))

    def searchhandler(self, searchstring, key):
        if key == curses.KEY_BACKSPACE:
            if self.searchpositions:
                self.items.selectbynr(self.searchpositions.pop())
        elif key in self.keybindings["repeatsearch"]:
            self.items.selectbyregexp(searchstring, includeselected=False)
        elif key == ord("\n"):
            self.searchpositions = []
            self.searchstring = searchstring
            hub.notify(events.activatefilelist())
        elif key == 1023:
            if self.searchpositions:
                self.items.selectbynr(self.searchpositions.pop())
            self.searchpositions = []
            self.searchstring = searchstring
            hub.notify(events.activatefilelist())
        else:
            self.searchpositions.append(self.items.selected)
            self.items.selectbyregexp(searchstring)
            # We explicitely issue a selectionchanged event because the
            # selectbyregexp doesn't do this due to the focus being on the
            # searchstring input window
        hub.notify(events.selectionchanged(self.items.getselected()))
        self.update()

    def focus_on_handler(self, searchstring, key):
        if key == ord("\n") and searchstring:
           self.items.focus_on(searchstring)

    def isclickonstring(self, y, x):
        """ check whether a click was on a string or not """
        while x < self.ix+self.iw:
            if self.win.inch(y, x) & 0xFF!=32:
                return 1
            x += 1
        return 0

    def resize(self, layout):
        h, w, y, x, self.border = layout
        window.window.resize(self, h, w, y, x)
        self.items._updatetop()

    # event handler

    def keypressed(self, event):
        if self.hasfocus():
            key = event.key
            if key in self.keybindings["selectnext"]:
                self.items.selectnext()
            elif key in self.keybindings["selectprev"]:
                self.items.selectprev()
            elif key in self.keybindings["selectnextpage"]:
                self.items.selectnextpage()
            elif key in self.keybindings["selectprevpage"]:
                self.items.selectprevpage()
            elif key in self.keybindings["selectfirst"]:
                self.items.selectfirst()
            elif key in self.keybindings["selectlast"]:
                self.items.selectlast()
            elif key in self.keybindings["dirdown"] and \
                     self.items.isdirselected():
                self.items.dirdown()
            elif key in self.keybindings["dirup"]:
                self.items.dirup()
            elif key in self.keybindings["addsongtoplaylist"] and \
                     self.items.issongselected():
                songtoadd = self.items.getselected()
                if self.items.selected is not self.lastadded:
                    self.lastadded = self.items.selected
                    hub.notify(events.playlistaddsongs([songtoadd]))
                    self.items.selectrelative(+1)
            elif key in self.keybindings["adddirtoplaylist"] and \
                     self.items.isdirselected():
                itemtoadd = self.items.getselected()
                if self.items.selected is not self.lastadded:
                    self.lastadded = self.items.selected
                    self.items.insertrecursiveselection()
                    self.items.selectrelative(+1)
            elif key in self.keybindings["playselectedsong"] and \
                     self.items.issongselected():
                songtoplay = self.items.getselected()
                hub.notify(events.playlistaddsongtop(songtoplay))
            elif key in self.keybindings["activateplaylist"]:
                hub.notify(events.activateplaylist())
            elif key in self.keybindings["insertrandomlist"] and self.items.isdirselected():
                self.items.randominsertrecursiveselection()
            elif key in self.keybindings["repeatsearch"]:
                if self.searchstring:
                    self.items.selectbyregexp(self.searchstring, includeselected=False)
            elif key in self.keybindings["search"]:
                hub.notify(events.requestinput(_("Search"),
                                                      "",
                                                      self.searchhandler))
            elif key in self.keybindings["focus"]:
                hub.notify(events.requestinput(_("Focus on"),
                                                 "",
                                                 self.focus_on_handler))
            elif key in self.keybindings["rescan"]:
                self.items.rescanselection(force=True)
                self.items.selectrelative(+1)
            elif key in self.keybindings["toggledelete"]:
                self.items.toggledeleteselection()
            elif ord("a")<=key-1024<=ord("z") or ord("A")<=key-1024<=ord("Z") :
                self.items.selectbyletter(chr(key-1024))
            elif ord("0")<=key<=ord("5"):
                if self.items.rateselection(key-ord("1")+1):
                    self.items.selectrelative(+1)
            else:
                return

            if self.items.selected != self.lastadded:
                self.lastadded = None

            self.update()
            raise hub.TerminateEventProcessing

    def mouseevent(self, event):
        if self.enclose(event.y, event.x):
            y, x = self.stdscrtowin(event.y, event.x)
            self.top()

            if event.state & curses.BUTTON1_CLICKED:
                if x==self.ix+self.iw and self.hasscrollbar:
                    scrollbarbegin, scrollbarheight = self.scrollbardimensions(self.items.top,
                                                                               len(self.items))
                    if y==self.iy+1:
                        self.items.selectprev()
                    elif y==self.iy+self.ih-2:
                        self.items.selectnext()
                    elif self.iy<y<scrollbarbegin:
                        self.items.selectprevpage()
                    elif scrollbarbegin+scrollbarheight<=y<self.iy+self.ih-2:
                        self.items.selectnextpage()
                elif self.items.selectbylinenumber(y-self.iy) and \
                   self.isclickonstring(y, x) and \
                   self.items.isdirselected():
                    self.items.dirdown()
            elif event.state & curses.BUTTON1_DOUBLE_CLICKED:
                if self.items.selectbylinenumber(y-self.iy) and \
                   self.isclickonstring(y, x):
                    if self.items.issongselected():
                        songtoadd = self.items.getselected()
                        self.lastadded = None
                        hub.notify(events.playlistaddsongs([songtoadd]))
                    else:
                        self.lastadded = None
                        self.items.insertrecursiveselection()
            elif event.state & curses.BUTTON3_CLICKED:
                self.items.dirup()
            else:
                return

            self.update()
            raise hub.TerminateEventProcessing

    def focuschanged(self, event):
        if self.hasfocus():
            self.lastadded = None
            hub.notify(events.selectionchanged(self.items.getselected()))
        self.update()

    # window update method

    def update(self):
        self.settitle(self.items.selectionpath())
        window.window.update(self)

        if self.hasfocus():
            self.updatestatusbar()

        showselectionbar = self.hasfocus() or self.searchpositions

        for i in range(self.items.top, self.items.top+self.ih):
            attr = curses.A_NORMAL
            if i<len(self.items):
                aitem = self.items[i]
                name = self.items[i].getname()
                if isinstance(aitem, item.song):
                    if i==self.items.selected and showselectionbar:
                        attr = self.colors.selected_song
                    else:
                        attr = self.colors.song
                elif isinstance(aitem, (item.artist, item.album)):
                    if i==self.items.selected and showselectionbar:
                        attr = self.colors.selected_artist_album
                    else:
                        attr = self.colors.artist_album
                else:
                    if i==self.items.selected and showselectionbar:
                        attr = self.colors.selected_directory
                    else:
                        attr = self.colors.directory
            else:
                name = ""
            name = encoding.encode(name)
            self.addstr(i-self.items.top+self.iy, self.ix, name.ljust(self.iw)[:self.iw], attr)

        self.updatescrollbar()

        # move cursor to the right position in order to make it more
        # easy for users of Braille displays to track the current
        # position/selection
        if self.hasfocus() and self.items.selected is not None:
            self.win.move(self.items.selected-self.items.top+1, 1)




# -*- coding: ISO-8859-1 -*-

# Copyright (C) 2003, 2004 Jörg Lehmann <joerg@luga.de>
#
# This file is part of PyTone (http://www.luga.de/pytone/)
#
# PyTone is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2
# as published by the Free Software Foundation.
#
# PyTone is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PyTone; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import exceptions
import threading
import Queue

import events, log

class TerminateEventProcessing(exceptions.Exception):
    pass

class DenyRequest(exceptions.Exception):
    """ deny processing of a request """
    pass

# from http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/87369

class PriorityQueue(Queue.Queue):

    def _init(self, maxsize):
        # we need to be sure to have a list as underlying queue
        self.maxsize = maxsize
        self.queue = []
        
    def _put(self, item):
	data, priority = item
	self._insort_right((priority, data))

    def _get(self):
	return self.queue.pop(0)[1]

    def _insort_right(self, x):
	"""Insert item x in list, and keep it sorted assuming a is sorted.

	If x is already in list, insert it to the right of the rightmost x.
	"""
	a = self.queue
	lo = 0
	hi = len(a)

	while lo < hi:
	    mid = (lo+hi)/2
	    if x[0] < a[mid][0]: hi = mid
	    else: lo = mid+1
	a.insert(lo, x)

#
# request response class
#

class requestresponse:
    """ structure containing request + response upon request """
    def __init__(self, request):
	self.request = request
	self.result = None
	self.ready = threading.Event()

    def __repr__(self):
	return "requestresponse(%r -> %r)" % (self.request, self.result)

    def waitforcompletion(self):
	self.ready.wait()

    def hascompleted(self):
	return self.ready.isSet()

#
# event and request dispatcher classes
#

class hub:

    """ collects event channels from different threads """

    def __init__(self):
	self.channels = []

    def connect(self, channel):
	self.channels.append(channel)

    def disconnect(self, channel):
	self.channels.remove(channel)

    def newchannel(self):
	achannel = channel(self)
	self.connect(achannel)
	return achannel

    def notify(self, item, priority=0):
	""" notify all channels belonging to hub of item (event or request) """
	log.debug("event: %s (priority %d)" % (repr(item), priority))
	for channel in self.channels:
	    channel._notify(item, priority)

    def request(self, request, priority=0):
	""" submit a request (blocking)

	this method submits a request, waits for the result and
	returns it.  Requests with a high priority are treated first.
	"""
	# generate a request response object for the request,
	# send it to hub and wait for result
	log.debug("request: %s (priority %d)" % (repr(request), priority))
	rr = requestresponse(request)
	self.notify(rr, priority)
	rr.waitforcompletion()
	return rr.result

#    def requestnoblocking(self, request, priority=0):
#        """ submit a request (nonblocking)
#
#        this method submits a request and returns a requestresponse
#        structure.  Requests with a high priority are treated first.
#        """
#        rr = requestresponse(request)
#        self.notify(rr, priority)
#        return rr


class channel:

    """ collects event handlers for one thread """

    def __init__(self, hub):
	self.hub = hub
	self.subscriptions = []
	self.suppliers = []
	self.queue = PriorityQueue(-1)
	# self.queue = Queue.Queue(-1)

    def process(self, block=False, timeout=None):
	""" process queued events and request

	If block is set, we wait for incoming events and requests.  In
        this case, a timeout in seconds can be specified, as well.
	"""

        while True:
            try:
                item = self.queue.get(block=block, timeout=timeout)
            except Queue.Empty:
                break
            # after having get the first event, we do no longer block
            block = False
            timeout = None
	    if isinstance(item, events.event):
		try:
		    for subscribedevent, handler in self.subscriptions:
			if isinstance(item, subscribedevent):
			    handler(item)
		except TerminateEventProcessing:
		    pass
	    else:
		for suppliedrequest, handler in self.suppliers:
		    if isinstance(item.request, suppliedrequest):
			# compute result and signalise that
			# request has been processed
			try:
			    item.result = handler(item.request)
			    log.debug(u"got result %r for %r" % (item.result, item.request))
			    item.ready.set()
			    break
			except DenyRequest:
			    pass

    def subscribe(self, eventtype, handler):
	self.subscriptions.append((eventtype, handler))

    def unsubscribe(self, eventtype, handler):
	self.subscriptions.remove((eventtype, handler))

    def supply(self, requesttype, handler):
	self.suppliers.append((requesttype, handler))

    def unsupply(self, requesttype, handler):
	self.suppliers.remove((requesttype, handler))

    def _notify(self, item, priority=0):
	""" notify channel of item (event or request) """
	self.queue.put((item, -priority))

# set up default hub and provide easy access to its externally used methods
_defaulthub = hub()

newchannel = _defaulthub.newchannel
notify = _defaulthub.notify
request = _defaulthub.request
